# -*- coding: utf-8 -*-
# Import Python libs
from __future__ import absolute_import

# Import PyLint libs
try:
    # >= pylint 1.0
    import astroid
except ImportError:  # pylint < 1.0
    from logilab import astng as astroid  # pylint: disable=no-name-in-module
from saltpylint.checkers import BaseChecker, utils

try:
    # >= pylint 1.0
    from pylint.interfaces import IAstroidChecker
except ImportError:  # < pylint 1.0
    from pylint.interfaces import IASTNGChecker as IAstroidChecker  # pylint: disable=no-name-in-module

try:
    from astroid.exceptions import NameInferenceError
except ImportError:
    class NameInferenceError(Exception):
        pass

WARNING_CODE = 'W1701'


class DunderDelChecker(BaseChecker):
    '''
    info: This class is used by pylint that checks
    if "__del__" is not used
    if "__del__" is used then a warning will be raised
    '''
    __implements__ = IAstroidChecker

    name = 'dunder-del'
    priority = -1
    msgs = {
        WARNING_CODE: (
            '"__del__" is not allowed!',
            'no-dunder-del',
            '"__del__" is not allowed! A "with" block could be a good solution'
        ),
    }

    def visit_functiondef(self, node):
        '''
        :param node: info about a function or method
        :return: None
        '''
        if node.name == "__del__":
            self.add_message(WARNING_CODE, node=node)


def register(linter):
    '''required method to auto register this checker '''
    linter.register_checker(DunderDelChecker(linter))



# -*- coding: utf-8 -*-
'''
    saltpylint.process
    ~~~~~~~~~~~~~~~~~~

    Lint checks to make sure subclasses of Salt's multiprocessing process implementations
    get it's parent run method called.
'''

# Import Python libs
from __future__ import absolute_import, unicode_literals, print_function

import astroid
from saltpylint.checkers import BaseChecker, utils
from pylint.interfaces import IAstroidChecker

MESSAGES = {
    'E3000': (
        'run method from base class %r is not called',
        'super-run-not-called',
        'Failed to call super(MySubClass, self).run()'
    )
}


def _ancestors_to_call(klass_node, method='run'):
    """return a dictionary where keys are the list of base classes providing
    the queried method, and so that should/may be called from the method node
    """
    to_call = {}
    for base_node in klass_node.ancestors(recurs=False):
        try:
            to_call[base_node] = next(base_node.igetattr(method))
        except astroid.InferenceError:
            continue
    return to_call


class ProcessSuperChecker(BaseChecker):

    __implements__ = IAstroidChecker
    name = 'process-super-checks'
    msgs = MESSAGES
    priority = -2

    def visit_functiondef(self, node):
        '''
        :param node: info about a function or method
        :return: None
        '''
        # ignore actual functions
        if not node.is_method():
            return

        if node.name != "run":
            return
        klass_node = node.parent.frame()
        import pprint
        #print('RUN', klass)
        #pprint.pprint(node.__dict__)
        to_call = _ancestors_to_call(klass_node)
        not_called_yet = dict(to_call)
        print('TO CALL')
        pprint.pprint(to_call)
        for stmt in node.nodes_of_class(astroid.Call):
            expr = stmt.func
            if not isinstance(expr, astroid.Attribute) or expr.attrname != 'run':
                continue
            # skip the test if using super
            print('EXPR', expr, expr.__dict__)
            if (
                isinstance(expr.expr, astroid.Call)
                and isinstance(expr.expr.func, astroid.Name)
                and expr.expr.func.name == "super"
            ):
                return
            try:
                for klass in expr.expr.infer():
                    if klass is astroid.Uninferable:
                        continue
                    # The inferred klass can be super(), which was
                    # assigned to a variable and the `run`
                    # was called later.
                    #
                    # base = super()
                    # base.run(...)

                    if (
                        isinstance(klass, astroid.Instance)
                        and isinstance(klass._proxied, astroid.ClassDef)
                        and is_builtin_object(klass._proxied)
                        and klass._proxied.name == "super"
                    ):
                        return
                    if isinstance(klass, objects.Super):
                        return
                    try:
                        del not_called_yet[klass]
                    except KeyError:
                        continue
                        #if klass not in to_call:
                        #    self.add_message(
                        #        "non-parent-init-called", node=expr, args=klass.name
                        #    )
            except astroid.InferenceError:
                continue
        for klass, method in not_called_yet.items():
            if utils.decorated_with(node, ["typing.overload"]):
                continue
            cls = utils.node_frame_class(method)
            if klass.name == "object" or (cls and cls.name == "object"):
                continue
            self.add_message('super-run-not-called', args=klass.name, node=node)

    def _visit_classdef(self, node):
        for base in node.bases:
            if not hasattr(base, 'attrname'):
                continue
            print('BASE', base.attrname)


def register(linter):
    '''
    Required method to auto register this checker
    '''
    linter.register_checker(ProcessSuperChecker(linter))



# -*- coding: utf-8 -*-
'''
    :codeauthor: :email:`Pedro Algarvio (pedro@algarvio.me)`
    :copyright: © 2017 by the SaltStack Team, see AUTHORS for more details.
    :license: Apache 2.0, see LICENSE for more details.


    saltpylint.thirdparty
    ~~~~~~~~~~~~~~~~~~~~~

    Checks all imports against a list of known and allowed 3rd-party modules
    and raises a lint error if an import not in that known 3rd-party modules list
    is not gated.
'''

# Import python libs
from __future__ import absolute_import
import os

# Import pylint libs
import astroid
import astroid.exceptions
import pkgutil
from astroid.modutils import is_relative, is_standard_module
from pylint.interfaces import IAstroidChecker
from saltpylint.checkers import BaseChecker, utils

MSGS = {
    'W8410': ('3rd-party module import is not gated in a try/except: %r',
              '3rd-party-module-not-gated',
              '3rd-party module imported without being gated in a try/except.'),
    'C8410': ('3rd-party local module import is not gated in a try/except: %r. '
              'Consider importing at the module global scope level and gate it '
              'in a try/except.',
              '3rd-party-local-module-not-gated',
              '3rd-party module locally imported without being gated. Consider importing '
              'at the module global scope level and gate it in a try/except'),
    }


def get_import_package(modname):
    '''
    Return the import package.

    Given modname is 'salt.utils', returns 'salt'
    '''
    return modname.split('.')[0]


class ThirdPartyImportsChecker(BaseChecker):

    __implements__ = IAstroidChecker

    name = '3rd-party-imports'
    msgs = MSGS
    priority = -2

    options = (
        ('allowed-3rd-party-modules', {
            'default': (),
            'type': 'csv',
            'metavar': '<3rd-party-modules>',
            'help': 'Known 3rd-party modules which don\' require being gated, separated by a comma'}),
    )

    known_py2_modules = ['__builtin__', 'exceptions']
    known_py3_modules = ['builtins']

    unix_modules = ('posix', 'pwd', 'spwd', 'grp', 'crypt', 'termios', 'tty', 'pty', 'fcntl', 'pipes', 'resource', 'nis', 'syslog', 'posixpath')
    win_modules = ('msilib', 'msvcrt', 'winreg', 'winsound', 'ntpath')

    all_modules = {m[1]: m[0].path for m in pkgutil.iter_modules()}
    std_modules_path = all_modules["os"]
    std_modules = []
    for mod, path in all_modules.items():
        if path == std_modules_path and mod not in unix_modules + win_modules:
            std_modules.append(mod)

    known_std_modules = known_py2_modules + known_py3_modules + std_modules

    def __init__(self, linter=None):
        BaseChecker.__init__(self, linter)
        self._inside_try_except = False
        self._inside_funcdef = False
        self._inside_if = False
        self.cwd = self.allowed_3rd_party_modules = None

    def open(self):
        super(ThirdPartyImportsChecker, self).open()
        self.cwd = os.getcwd()
        self.allowed_3rd_party_modules = set(self.config.allowed_3rd_party_modules)  # pylint: disable=no-member

    # pylint: disable=unused-argument
    @utils.check_messages('3rd-party-imports')
    def visit_if(self, node):
        self._inside_if = True

    @utils.check_messages('3rd-party-imports')
    def leave_if(self, node):
        self._inside_if = True

    @utils.check_messages('3rd-party-imports')
    def visit_tryexcept(self, node):
        self._inside_try_except = True

    @utils.check_messages('3rd-party-imports')
    def leave_tryexcept(self, node):
        self._inside_try_except = False

    @utils.check_messages('3rd-party-imports')
    def visit_functiondef(self, node):
        self._inside_funcdef = True

    @utils.check_messages('3rd-party-imports')
    def leave_functiondef(self, node):
        self._inside_funcdef = False
    # pylint: enable=unused-argument

    @utils.check_messages('3rd-party-imports')
    def visit_import(self, node):
        names = [name for name, _ in node.names]
        for name in names:
            self._check_third_party_import(node, name)

    @utils.check_messages('3rd-party-imports')
    def visit_importfrom(self, node):
        self._check_third_party_import(node, node.modname)

    def _check_third_party_import(self, node, modname):
        if modname in self.known_std_modules:
            # Don't even care about these
            return
        module_file = node.root().file

        if is_relative(modname, module_file):
            # Is the import relative to the curent module being checked
            return

        base_modname = modname.split('.', 1)[0]
        import_modname = modname
        while True:
            try:
                imported_module = node.do_import_module(import_modname)
                if not imported_module:
                    break
                if imported_module.file.startswith(self.cwd):
                    # This is an import to package under the project being tested
                    return
                # If we reached this far, we were able to import the module but it's
                # not considered a module from within the project being checked
                break
            except Exception:  # pylint: disable=broad-except
                # This is, for example, from salt.ext.six.moves import Y
                # Because `moves` is a dynamic/runtime module
                import_modname = import_modname.rsplit('.', 1)[0]

            if import_modname == base_modname or not import_modname:
                break

        try:
            if not is_standard_module(modname):
                if self._inside_try_except is False:
                    if get_import_package(modname) in self.allowed_3rd_party_modules:
                        return
                    if self._inside_if or self._inside_funcdef:
                        message_id = '3rd-party-local-module-not-gated'
                    else:
                        message_id = '3rd-party-module-not-gated'
                    self.add_message(message_id, node=node, args=modname)
        except astroid.exceptions.AstroidBuildingException:
            # Failed to import
            if self._inside_try_except is False:
                if get_import_package(modname) in self.allowed_3rd_party_modules:
                    return
                if self._inside_if or self._inside_funcdef:
                    message_id = '3rd-party-local-module-not-gated'
                else:
                    message_id = '3rd-party-module-not-gated'
                self.add_message(message_id, node=node, args=modname)
        except astroid.exceptions.InferenceError:
            # Failed to import
            if self._inside_try_except is False:
                if get_import_package(modname) in self.allowed_3rd_party_modules:
                    return
                if self._inside_if or self._inside_funcdef:
                    message_id = '3rd-party-local-module-not-gated'
                else:
                    message_id = '3rd-party-module-not-gated'
                self.add_message(message_id, node=node, args=modname)
        except ImportError:
            # Definitly not a standard library import
            if self._inside_try_except is False:
                if get_import_package(modname) in self.allowed_3rd_party_modules:
                    return
                if self._inside_if or self._inside_funcdef:
                    message_id = '3rd-party-local-module-not-gated'
                else:
                    message_id = '3rd-party-module-not-gated'
                self.add_message(message_id, node=node, args=modname)


def register(linter):
    '''required method to auto register this checker '''
    linter.register_checker(ThirdPartyImportsChecker(linter))



# -*- coding: utf-8 -*-
'''
    saltpylint.checkers
    ~~~~~~~~~~~~~~~~~~~~

    Works around older astroid versions
'''
# Import python libs
from __future__ import absolute_import

# Import pylint libs
import astroid
from pylint.checkers import BaseChecker as _BaseChecker
# Imported to avoid needing a separate import from pylint.checkers
from pylint.checkers import utils


class BaseChecker(_BaseChecker):
    def __init__(self, *args, **kwargs):
        super(BaseChecker, self).__init__(*args, **kwargs)
        if hasattr(self, 'visit_call') and not hasattr(astroid, 'Call'):
            setattr(self, 'visit_callfunc', self.visit_call)




from __future__ import absolute_import
import astroid

from pylint.interfaces import IAstroidChecker
from pylint.checkers import BaseChecker

class VirtChecker(BaseChecker):
    '''
    checks for compliance inside __virtual__
    '''
    __implements__ = IAstroidChecker

    name = 'virt-checker'

    VIRT_LOG = 'log-in-virtual'

    msgs = {
        'E1401': ('Log statement detected inside __virtual__ function. Remove it.',
                  VIRT_LOG,
                  'Loader processes __virtual__ so logging not in scope'),
    }
    options = ()

    priority = -1

    def visit_functiondef(self, node):
        '''
        Verifies no logger statements inside __virtual__
        '''
        if (not isinstance(node, astroid.FunctionDef) or
            node.is_method()
            or node.type != 'function'
            or not node.body
           ):
            # only process functions
            return

        try:
            if not node.name == '__virtual__':
                # only need to process the __virtual__ function
                return
        except AttributeError:
            return

        # walk contents of __virtual__ function
        for child in node.get_children():
            for functions in child.get_children():
                if isinstance(functions, astroid.Call):
                    if isinstance(functions.func, astroid.Attribute):
                        try:
                            # Inspect each statement for an instance of 'logging'
                            for inferred in functions.func.expr.infer():
                                try:
                                    instance_type = inferred.pytype().split('.')[0]
                                except TypeError:
                                    continue
                                if instance_type == 'logging':
                                    self.add_message(
                                        self.VIRT_LOG, node=functions
                                    )
                                    # Found logger, don't need to keep processing this line
                                    break
                        except AttributeError:
                            # Not a log function
                            return

def register(linter):
    '''
    required method to auto register this checker
    '''
    linter.register_checker(VirtChecker(linter))
